@@ -11,7 +11,8 @@ import atexit
 import code
 import signal
 import errno
-import platform
+
+__version__ = '1.0.0'
 
 try:
     import signalfd
@@ -59,11 +60,11 @@ except ImportError:  # python 3
 _ORIGINAL_THREAD = _get_original('threading.Thread')
 _ORIGINAL_EVENT = _get_original('threading.Event')
 _ORIGINAL__ACTIVE = _get_original('threading._active')
+_ORIGINAL_SLEEP = _get_original('time.sleep')
 
 PY3 = sys.version_info[0] == 3
 PY26 = sys.version_info[:2] == (2, 6)
 VERBOSE = True
-START_TIMEOUT = None
 
 try:
     import ctypes
@@ -81,35 +82,31 @@ try:
 except ImportError:
     pthread_setname_np = lambda ident, name: None
 
-# OS X getsockopt(2) defines (may work for BSD too?)
-SOL_LOCAL = 0
-LOCAL_PEERCRED = 1
 
-SO_PEERCRED = 17
-
-
-def cry(message):
+def cry(message, time=_get_original('time.time')):
     """
     Fail-ignorant logging function.
     """
     if VERBOSE:
         try:
-            _STDERR.write("Manhole: %s\n" % message)
+            _STDERR.write("Manhole[%.4f]: %s\n" % (time(), message))
         except:  # pylint: disable=W0702
             pass
 
 
+if sys.platform == 'darwin' or sys.platform.startswith("freebsd"):
+    _PEERCRED_LEVEL = getattr(socket, 'SOL_LOCAL', 0)
+    _PEERCRED_OPTION = getattr(socket, 'LOCAL_PEERCRED', 1)
+else:
+    _PEERCRED_LEVEL = socket.SOL_SOCKET
+    # TODO: Is this missing on some platforms?
+    _PEERCRED_OPTION = getattr(socket, 'SO_PEERCRED', 17)
+
+
 def get_peercred(sock):
     """Gets the (pid, uid, gid) for the client on the given *connected* socket."""
-
-    if platform.system() == 'Darwin':
-        return struct.unpack('3i', sock.getsockopt(
-            SOL_LOCAL, LOCAL_PEERCRED, struct.calcsize('3i')
-        ))
-    else:
-        return struct.unpack('3i', sock.getsockopt(
-            socket.SOL_SOCKET, SO_PEERCRED, struct.calcsize('3i')
-        ))
+    buf = sock.getsockopt(_PEERCRED_LEVEL, _PEERCRED_OPTION, struct.calcsize('3i'))
+    return struct.unpack('3i', buf)
 
 
 class SuspiciousClient(Exception):
@@ -118,17 +115,41 @@ class SuspiciousClient(Exception):
 
 class Manhole(_ORIGINAL_THREAD):
     """
-    Thread that runs the infamous "Manhole".
+    Thread that runs the infamous "Manhole". This thread is a `daemon` thread - it will exit if the main thread
+    exits.
+
+    On connect, a different, non-daemon thread will be started - so that the process won't exit while there's a
+    connection to the manole.
+
+    Args:
+        sigmask (list of singal numbers): Signals to block in this thread.
+        start_timeout (float): Seconds to wait for the thread to start. Emits a message if the thread is not running
+            when calling ``start()``.
+        bind_delay (float): Seconds to delay socket binding. Default: `no delay`.
+        daemon_connection (bool): The connection thread is daemonic (dies on app exit). Default: ``False``.
     """
 
-    def __init__(self, sigmask, start_timeout):
+    def __init__(self, sigmask, start_timeout, bind_delay=None, locals=None, daemon_connection=False):
         super(Manhole, self).__init__()
         self.daemon = True
+        self.daemon_connection = daemon_connection
         self.name = "Manhole"
         self.sigmask = sigmask
         self.serious = _ORIGINAL_EVENT()
-        self.start_timeout = start_timeout  # time to wait for the manhole to get serious (to have a complete start)
-                                            # see: http://emptysqua.re/blog/dawn-of-the-thread/
+        # time to wait for the manhole to get serious (to have a complete start)
+        # see: http://emptysqua.re/blog/dawn-of-the-thread/
+        self.start_timeout = start_timeout
+        self.bind_delay = bind_delay
+        self.locals = locals
+
+    def clone(self, **kwargs):
+        """
+        Make a fresh thread with the same options. This is usually used on dead threads.
+        """
+        return Manhole(
+            self.sigmask, self.start_timeout, locals=self.locals, daemon_connection=self.daemon_connection,
+            **kwargs
+        )
 
     def start(self):
         super(Manhole, self).start()
@@ -138,26 +159,35 @@ class Manhole(_ORIGINAL_THREAD):
     @staticmethod
     def get_socket():
         sock = _ORIGINAL_SOCKET(socket.AF_UNIX, socket.SOCK_STREAM)
-        pid = os.getpid()
-        name = "/tmp/manhole-%s" % pid
+        name = _manhole_uds_name()
         if os.path.exists(name):
             os.unlink(name)
         sock.bind(name)
         sock.listen(5)
         cry("Manhole UDS path: "+name)
-        return sock, pid
+        return sock
 
     def run(self):
+        """
+        Runs the manhole loop. Only accepts one connection at a time because:
+
+        * This thread is a daemon thread (exits when main thread exists).
+        * The connection need exclusive access to stdin, stderr and stdout so it can redirect inputs and outputs.
+        """
         self.serious.set()
         if signalfd and self.sigmask:
             signalfd.sigprocmask(signalfd.SIG_BLOCK, self.sigmask)
         pthread_setname_np(self.ident, self.name)
 
-        sock, pid = self.get_socket()
+        if self.bind_delay:
+            cry("Delaying UDS binding %s seconds ..." % self.bind_delay)
+            _ORIGINAL_SLEEP(self.bind_delay)
+
+        sock = self.get_socket()
         while True:
-            cry("Waiting for new connection (in pid:%s) ..." % pid)
+            cry("Waiting for new connection (in pid:%s) ..." % os.getpid())
             try:
-                client = ManholeConnection(sock.accept()[0], self.sigmask)
+                client = ManholeConnection(sock.accept()[0], self.locals, self.daemon_connection)
                 client.start()
                 client.join()
             except (InterruptedError, socket.error) as e:
@@ -169,25 +199,29 @@ class Manhole(_ORIGINAL_THREAD):
 
 
 class ManholeConnection(_ORIGINAL_THREAD):
-    def __init__(self, client, sigmask):
+    """
+    Manhole thread that handles the connection. This thread is a normal thread (non-daemon) - it won't exit if the
+    main thread exits.
+    """
+    def __init__(self, client, locals, daemon=False):
         super(ManholeConnection, self).__init__()
-        self.daemon = False
+        self.daemon = daemon
         self.client = client
         self.name = "ManholeConnection"
-        self.sigmask = sigmask
+        self.locals = locals
 
     def run(self):
         cry('Started ManholeConnection thread. Checking credentials ...')
-        if signalfd and self.sigmask:
-            signalfd.sigprocmask(signalfd.SIG_BLOCK, self.sigmask)
         pthread_setname_np(self.ident, "Manhole ----")
-
         pid, _, _ = self.check_credentials(self.client)
         pthread_setname_np(self.ident, "Manhole %s" % pid)
-        self.handle(self.client)
+        self.handle(self.client, self.locals)
 
     @staticmethod
     def check_credentials(client):
+        """
+        Checks credentials for given socket.
+        """
         pid, uid, gid = get_peercred(client)
 
         euid = os.geteuid()
@@ -201,40 +235,37 @@ class ManholeConnection(_ORIGINAL_THREAD):
         return pid, uid, gid
 
     @staticmethod
-    def handle(client):
+    def handle(client, locals):
+        """
+        Handles connection. This is a static method so it can be used without a thread (eg: from a signal handler -
+        `oneshot_on`).
+        """
         client.settimeout(None)
 
-        # disable this till we have evidence that it's needed
-        #client.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 0)
-        # Note: setting SO_RCVBUF on UDS has no effect, see: http://man7.org/linux/man-pages/man7/unix.7.html
+        # # disable this till we have evidence that it's needed
+        # client.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 0)
+        # # Note: setting SO_RCVBUF on UDS has no effect, see: http://man7.org/linux/man-pages/man7/unix.7.html
 
         backup = []
         old_interval = getinterval()
+        patches = [('r', ('stdin', '__stdin__')), ('w', ('stdout', '__stdout__'))]
+        if _REDIRECT_STDERR:
+            patches.append(('w', ('stderr', '__stderr__')))
         try:
             try:
                 client_fd = client.fileno()
-                for mode, names in (
-                    ('w', (
-                        'stderr',
-                        'stdout',
-                        '__stderr__',
-                        '__stdout__'
-                    )),
-                    ('r', (
-                        'stdin',
-                        '__stdin__'
-                    ))
-                ):
+                for mode, names in patches:
                     for name in names:
                         backup.append((name, getattr(sys, name)))
                         setattr(sys, name, _ORIGINAL_FDOPEN(client_fd, mode, 1 if PY3 else 0))
-                run_repl()
+                run_repl(locals)
                 cry("DONE.")
             finally:
                 try:
-                    setinterval(2147483647)  # change the switch/check interval to something ridiculous
-                                             # we don't want to have other thread try to write to the
-                                             # redirected sys.__std*/sys.std* - it would fail horribly
+                    # Change the switch/check interval to something ridiculous. We don't want to have other thread try
+                    # to write to the redirected sys.__std*/sys.std* - it would fail horribly.
+                    setinterval(2147483647)
+
                     client.close()  # close before it's too late. it may already be dead
                     junk = []  # keep the old file objects alive for a bit
                     for name, fh in backup:
@@ -256,24 +287,44 @@ class ManholeConnection(_ORIGINAL_THREAD):
             cry(traceback.format_exc())
 
 
-def run_repl():
+class ManholeConsole(code.InteractiveConsole):
+
+    def __init__(self, *args, **kw):
+        code.InteractiveConsole.__init__(self, *args, **kw)
+        if _REDIRECT_STDERR:
+            self.file = sys.stderr
+        else:
+            self.file = sys.stdout
+
+    def write(self, data):
+        self.file.write(data)
+
+
+def run_repl(locals):
+    """
+    Dumps stacktraces and runs an interactive prompt (REPL).
+    """
     dump_stacktraces()
-    code.InteractiveConsole({
+    namespace = {
         'dump_stacktraces': dump_stacktraces,
         'sys': sys,
         'os': os,
         'socket': socket,
         'traceback': traceback,
-    }).interact()
+    }
+    if locals:
+        namespace.update(locals)
+    ManholeConsole(namespace).interact()
 
 
 def _handle_oneshot(_signum, _frame):
+    assert _INST, "Manhole wasn't installed !"
     try:
-        sock, pid = Manhole.get_socket()
-        cry("Waiting for new connection (in pid:%s) ..." % pid)
+        sock = Manhole.get_socket()
+        cry("Waiting for new connection (in pid:%s) ..." % os.getpid())
         client, _ = sock.accept()
         ManholeConnection.check_credentials(client)
-        ManholeConnection.handle(client)
+        ManholeConnection.handle(client, _INST.locals)
     except:  # pylint: disable=W0702
         # we don't want to let any exception out, it might make the application missbehave
         cry("Manhole oneshot connection failed:")
@@ -283,12 +334,22 @@ def _handle_oneshot(_signum, _frame):
 
 
 def _remove_manhole_uds():
-    name = "/tmp/manhole-%s" % os.getpid()
+    name = _manhole_uds_name()
     if os.path.exists(name):
         os.unlink(name)
 
+
+def _manhole_uds_name():
+    if _SOCKET_PATH is None:
+        return "/tmp/manhole-%s" % os.getpid()
+    return _SOCKET_PATH
+
+
 _INST_LOCK = _ORIGINAL_ALLOCATE_LOCK()
 _STDERR = _INST = _ORIGINAL_OS_FORK = _ORIGINAL_OS_FORKPTY = _SHOULD_RESTART = None
+_SOCKET_PATH = None
+_REINSTALL_DELAY = None
+_REDIRECT_STDERR = True
 
 
 def _patched_fork():
@@ -328,14 +389,42 @@ ALL_SIGNALS = [
 ]
 
 
-def install(verbose=True, patch_fork=True, activate_on=None, sigmask=ALL_SIGNALS, oneshot_on=None, start_timeout=0.5):
-    global _STDERR, _INST, _SHOULD_RESTART, VERBOSE, START_TIMEOUT  # pylint: disable=W0603
+def install(verbose=True, patch_fork=True, activate_on=None, sigmask=ALL_SIGNALS, oneshot_on=None, start_timeout=0.5,
+            socket_path=None, reinstall_delay=0.5, locals=None, daemon_connection=False, redirect_stderr=True):
+    """
+    Installs the manhole.
+
+    Args:
+        verbose (bool): Set it to ``False`` to squelch the stderr ouput
+        patch_fork (bool): set it to ``False`` if you don't want your ``os.fork`` and ``os.forkpy`` monkeypatched
+        activate_on (int or signal name): set to ``"USR1"``, ``"USR2"`` or some other signal name, or a number if you
+            want the Manhole thread to start when this signal is sent. This is desireable in case you don't want the
+            thread active all the time.
+        oneshot_on (int or signal name): set to ``"USR1"``, ``"USR2"`` or some other signal name, or a number if you
+            want the Manhole to listen for connection in the signal handler. This is desireable in case you don't want
+            threads at all.
+        sigmask (list of ints or signal names): will set the signal mask to the given list (using
+            ``signalfd.sigprocmask``). No action is done if ``signalfd`` is not importable.
+            **NOTE**: This is done so that the Manhole thread doesn't *steal* any signals; Normally that is fine cause
+            Python will force all the signal handling to be run in the main thread but signalfd doesn't.
+        socket_path (str): Use a specifc path for the unix domain socket (instead of ``/tmp/manhole-<pid>``). This
+            disables ``patch_fork`` as children cannot resuse the same path.
+        reinstall_delay (float): Delay the unix domain socket creation *reinstall_delay* seconds. This
+            alleviates cleanup failures when using fork+exec patterns.
+        locals (dict): Names to add to manhole interactive shell locals.
+        daemon_connection (bool): The connection thread is daemonic (dies on app exit). Default: ``False``.
+        redirect_stderr (bool): Redirect output from stderr to manhole console. Default: ``True``.
+    """
+    global _STDERR, _INST, _SHOULD_RESTART  # pylint: disable=W0603
+    global VERBOSE, _REINSTALL_DELAY, _SOCKET_PATH, _REDIRECT_STDERR  # pylint: disable=W0603
     with _INST_LOCK:
         VERBOSE = verbose
-        START_TIMEOUT = start_timeout
+        _SOCKET_PATH = socket_path
+        _REINSTALL_DELAY = reinstall_delay
+        _REDIRECT_STDERR = redirect_stderr
         _STDERR = sys.__stderr__
         if not _INST:
-            _INST = Manhole(sigmask, start_timeout)
+            _INST = Manhole(sigmask, start_timeout, locals=locals, daemon_connection=daemon_connection)
             if oneshot_on is not None:
                 oneshot_on = getattr(signal, 'SIG'+oneshot_on) if isinstance(oneshot_on, string) else oneshot_on
                 signal.signal(oneshot_on, _handle_oneshot)
@@ -347,30 +436,39 @@ def install(verbose=True, patch_fork=True, activate_on=None, sigmask=ALL_SIGNALS
             else:
                 activate_on = getattr(signal, 'SIG'+activate_on) if isinstance(activate_on, string) else activate_on
                 if activate_on == oneshot_on:
-                    raise RuntimeError('You cannot do activation of the Manhole thread on the same signal that you want to do oneshot activation !')
+                    raise RuntimeError('You cannot do activation of the Manhole thread on the same signal '
+                                       'that you want to do oneshot activation !')
                 signal.signal(activate_on, _activate_on_signal)
         atexit.register(_remove_manhole_uds)
         if patch_fork:
-            if activate_on is None and oneshot_on is None:
+            if activate_on is None and oneshot_on is None and socket_path is None:
                 _patch_os_fork_functions()
             else:
                 if activate_on:
                     cry("Not patching os.fork and os.forkpty. Activation is done by signal %s" % activate_on)
                 elif oneshot_on:
                     cry("Not patching os.fork and os.forkpty. Oneshot activation is done by signal %s" % oneshot_on)
+                elif socket_path:
+                    cry("Not patching os.fork and os.forkpty. Using user socket path %s" % socket_path)
 
 
 def reinstall():
+    """
+    Reinstalls the manhole. Checks if the thread is running. If not, it starts it again.
+    """
     global _INST  # pylint: disable=W0603
     assert _INST
     with _INST_LOCK:
         if not (_INST.is_alive() and _INST in _ORIGINAL__ACTIVE):
-            _INST = Manhole(_INST.sigmask, START_TIMEOUT)
+            _INST = _INST.clone(bind_delay=_REINSTALL_DELAY)
             if _SHOULD_RESTART:
                 _INST.start()
 
 
 def dump_stacktraces():
+    """
+    Dumps thread ids and tracebacks to stdout.
+    """
     lines = []
     for thread_id, stack in sys._current_frames().items():  # pylint: disable=W0212
         lines.append("\n######### ProcessID=%s, ThreadID=%s #########" % (
@@ -382,4 +480,4 @@ def dump_stacktraces():
                 lines.append("  %s" % (line.strip()))
     lines.append("#############################################\n\n")
 
-    print('\n'.join(lines), file=sys.stderr)
+    print('\n'.join(lines), file=sys.stderr if _REDIRECT_STDERR else sys.stdout)
